#include "arraysize.h"
#include "arraystate.h"
#include "bufferstate.h"
#include "enums.h"
#include "strformat.h"

using namespace std;

BufferContent::BufferContent():
	consistent(true),
	stride(0)
{ }

void BufferContent::update(const ArrayState &array)
{
	if(array.stride!=stride)
		consistent = false;

	stride = array.stride;

	for(vector<Array>::iterator i=arrays.begin(); i!=arrays.end(); ++i)
		if(i->kind==array.kind && i->index==array.index)
		{
			if(array.size!=i->size || array.type!=i->type || array.pointer!=i->offset)
			{
				consistent = false;
				arrays.erase(i);
			}
			else
				return;


			break;
		}

	vector<Array>::iterator place = arrays.end();
	for(vector<Array>::iterator i=arrays.begin(); i!=arrays.end(); ++i)
		if(i->offset>array.pointer)
		{
			place = i;
			break;
		}

	arrays.insert(place, array);
}

void BufferContent::update_elements(GLenum type)
{
	if(arrays.empty())
	{
		Array array;
		array.kind = GL_ELEMENT_ARRAY_BUFFER;
		array.type = type;
		arrays.push_back(array);
		stride = typesize(type);
	}
	else if(arrays.size()>1 || arrays.front().kind!=GL_ELEMENT_ARRAY_BUFFER)
		consistent = false;
	else
	{
		if(arrays.front().type!=type)
			consistent = false;
		arrays.front().type = type;
		stride = typesize(type);
	}
}

string BufferContent::describe() const
{
	if(arrays.empty())
		return "(unknown)";

	string result;
	for(vector<Array>::const_iterator i=arrays.begin(); i!=arrays.end(); ++i)
	{
		char kind = '?';
		if(i->kind==GL_VERTEX_ARRAY)
			kind = 'V';
		else if(i->kind==GL_NORMAL_ARRAY)
			kind = 'N';
		else if(i->kind==GL_COLOR_ARRAY)
			kind = 'C';
		else if(i->kind==GL_TEXTURE_COORD_ARRAY && i->index==0)
			kind = 'T';
		else if(i->kind==GL_ELEMENT_ARRAY_BUFFER)
			kind = 'E';

		char type[3] = { '?', 0, 0 };
		if(i->type==GL_FLOAT)
			type[0] = 'F';
		else if(i->type==GL_DOUBLE)
			type[0] = 'D';
		else if(i->type==GL_INT || i->type==GL_UNSIGNED_INT)
			type[0] = 'I';
		else if(i->type==GL_SHORT || i->type==GL_UNSIGNED_SHORT)
			type[0] = 'S';
		else if(i->type==GL_BYTE || i->type==GL_UNSIGNED_BYTE)
			type[0] = 'B';

		if(i->type==GL_UNSIGNED_INT || i->type==GL_UNSIGNED_SHORT || i->type==GL_UNSIGNED_BYTE)
		{
			type[1] = type[0];
			type[0] = 'U';
		}

		if(!result.empty())
			result += '_';
		result += strformat("%c%d%s", kind, i->size, type);
	}

	return result;
}


BufferContent::Array::Array():
	kind(GL_NONE),
	index(0),
	size(1),
	type(GL_NONE),
	offset(0)
{ }

BufferContent::Array::Array(const ArrayState &a):
	kind(a.kind),
	index(a.index),
	size(a.size),
	type(a.type),
	offset(a.pointer)
{ }


BufferState::BufferState():
	id(0),
	target(0),
	usage(GL_STATIC_DRAW),
	size(0),
	data(0)
{ }

BufferState::~BufferState()
{
	delete[] data;
}

void BufferState::set_data(unsigned sz, const void *ptr, GLenum use)
{
	usage = use;
	size = sz;
	delete[] data;
	data = new char[size];
	if(ptr)
		set_sub_data(0, size, ptr);
	content = BufferContent();
}

void BufferState::set_sub_data(unsigned off, unsigned sz, const void *ptr)
{
	if(data && off+sz<=size)
	{
		const char *cptr = reinterpret_cast<const char *>(ptr);
		copy(cptr, cptr+sz, data+off);
	}
}

string BufferState::describe() const
{
	if(content.stride)
	{
		const char *what = (content.arrays.front().kind==GL_ELEMENT_ARRAY_BUFFER ? "indices" : "vertices");
		return content.describe()+strformat(", %d %s (%d bytes), %s",
			size/content.stride, what, size, describe_enum(usage, ""));
	}
	else
		return strformat("%d bytes, %s", size, describe_enum(usage, ""));
}


BufferBindingState::BufferBindingState():
	buffer(0),
	offset(0),
	size(0)
{ }
